//
// MarbleClock.c (c) 2008, Mats Engstrom
// Version 1.0
//
// See http://www.instructables.com/id/E5GQP1GFCEILJ2E/ for more information
//
// This software is licenced as: 
//	Attribution Non-commercial Share Alike (by-nc-sa)
//
// This license lets others remix, tweak, and build upon your work non-commercially, 
// as long as they credit you and license their new creations under the identical terms. 
// Others can download and redistribute your work just like the by-nc-nd license, but 
// they can also translate, make remixes, and produce new stories based on your work. 
// All new work based on yours will carry the same license, so any derivatives will 
// also be non-commercial in nature. 

#define F_CPU 10000000UL

#include <inttypes.h>
#include <util/delay.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <avr/sleep.h>


/*
         Tiny2313
         +--------------------------------------------+
   reset |  1 PA2/-reset                       VCC 20 | +5 volt
  led m1 |  2 PD0/rxd            int7/scl/usck/PB7 19 | sck   
  led m2 |  3 PD1/txd             int6/do/miso/PB6 18 | miso  
   xtal2 |  4 PA1/xtal2           int5/di/mosi/PB5 17 | mosi  
   xtal1 |  5 PA0/xtal1              int4/oc1b/PB4 16 | led h16
  led m4 |  6 PD2/int0/ckout         int3/oc1a/PB3 15 | led h8
  led m8 |  7 PD3/int1               int2/oc0a/PB2 14 | led h4
 led m16 |  8 PD4/t0                 int1/ain1/PB1 13 | led h2
 led m32 |  9 PD5/t1/oc0b            int0/ain0/PB0 12 | led h1
     gnd | 10 GND                          icp/PD6 11 | button
        +--------------------------------------------+
*/ 


#define BUTTON			PD6		// The time set button

volatile uint16_t	ms100=0;	// ISR incremets this at 100 mS interval
volatile uint8_t	second=0;	// The current time - seconds
volatile uint8_t	minute=0;	// The current time - minutes
volatile uint8_t	hour=12;	// The current time - hours


void InitCPU(void);
void HandleButtonPressed(uint16_t *presstime);
void HandleButtonReleased(uint16_t *presstime);
void AttractMode();

//
// This Interupt Serivce Routine gets called every 1.6384 mS by timer 0
//
ISR (SIG_OVERFLOW0) { 
static int32_t	ticks = 0;
static int8_t	ms100ticks = 0;

	// Standard Bresenhams time-keeping algorithm
	ticks-=16384;
	if (ticks<0) {
		ticks+=10000000;
		
		// Incremet time by one second
		second++;
		if (second>=60) {
			second=0;
			minute++;
			if (minute>=60) {
				minute=0;
				hour++;
				if (hour>=24) {
					hour=0;
				}
			}
		}
	}

	// If this ISR been executed 61 times (approx 100mS)
	// then increment the 100mS-counter
	ms100ticks++;
	if (ms100ticks>61) {
		ms100ticks=0;
		ms100++;
	}

}




//
// The main function.
// Initializes registers, then loops forever updating the
// leds and checks the status of the time set button.
// 
int main (void) {
	uint16_t	presstime=0;

	// Initialize timer and ports
	InitCPU();

	// Flash the leds to signal loss of time
	AttractMode();

	// Set the time to noon after power on
	hour = 12;
	minute = 0;
	second = 0;

	// Loop forever
	for (;;) {
		// Update the leds 
		PORTD = minute | _BV(BUTTON);
		PORTB = hour;

		// Check whether the button is pressed or released
		if ((PIND & _BV(BUTTON)) == 0) {
			HandleButtonPressed(&presstime);
		} else {
			HandleButtonReleased(&presstime);
		}
	}
}

	

//
// Initialize and start the timer (timer 0) responsible for keeping
// track of the time. Also sets the ports where the leds are connected
// to as outputs.
//
void InitCPU(void) {
	TIMSK = _BV(TOIE0);   			//Enable TCNT0 overflow
	TCNT0 = 0;         				//Reset TCNT0
	TCCR0A = 0;						//Normal overflow
	TCCR0B = _BV(CS00) |_BV(CS01);  //Prescale /64

	sei();							// Globally enable interrupts

	DDRD = 0b00111111;				// Set PD0-5 (minutes) as outputs
	DDRB = 0b00011111;				// Set PB0-4 (hours) as outputs
	PORTD = _BV(BUTTON);			// Enable pullup for the button
}




//
// This function gets called when the button is pressed.
// It incremets the time slowly between 0.6 and 15 seconds of pressing,
// after 15 seconds it goes faster.
//
// The reason for not starting until after 0.6 seconds is because a short
// press will be handled by the HandleButtonReleased() as a decrement 
// of the time by one second.
//
void HandleButtonPressed(uint16_t *presstime) {
	static uint16_t	lastms100 = 0;

	// Return if we're in same 100 mS-tick as last time called
	if (lastms100 == ms100) return;	
	lastms100 = ms100;

	// Update time pressed (100 mS incremets)
	(*presstime)++;

	// Return if button been pressed less than 0.6 seconds  
	if (*presstime < 6) return;		
	

	// If the button been pressed less than 15 seconds increment time slowly
	// If the button been pressed more than 15 seconds increment time fast
	if (
		(((*presstime) < 150) && (((*presstime) % 6) == 0))
		||
		((*presstime) > 150)
	   ) {
		// Increment time by one minute, always set seconds to zero
		second = 0;
		minute++;
		if (minute >= 60) {
			minute = 0;
			hour++;
			if (hour >= 24) {
				hour = 0;
			}
		}
	}
}



//
// This function is called when the button is not pressed.
// If it's been pressed for a short time then decrement the time.
//
void HandleButtonReleased(uint16_t *presstime) {
	// If the press was between 0.2 and 0.5 seconds long
	// then decrement one minute and set seconds to zero
	if ((*presstime > 1) && (*presstime < 6)) {
		second = 0;
		if (minute > 0) {
			minute--;
		} else {
			minute = 59;
			if (hour > 0) {
				hour--;
			} else {
				hour = 23;
			}
		}
	}

	// The button is released, reset the time to zero
	*presstime = 0;
}


//
// The attract mode show a "nightrider" display until
// the user presses the time set button.
//
void AttractMode() {
	uint8_t	ledpos = 0;
	uint8_t	direction = 1;
	uint8_t	hbits;
	uint8_t	mbits;
	uint16_t lastms100;	
		
	// Repeat until key pressed
	while ((PIND & _BV(BUTTON)) != 0) {
		// Wait 100 mS
		while (ms100 == lastms100);
		lastms100 = ms100;

		// Light the led pointed to by ledpos
		hbits = 0;
		mbits = 0;
		if (ledpos < 6) mbits = _BV(ledpos);
		if (ledpos >= 6) hbits = _BV(ledpos-6);
		PORTD = mbits | _BV(BUTTON);
		PORTB = hbits;

		// Move to the next led
		ledpos = ledpos + direction;
		// Reverse direction if reached the first or last led
		if (ledpos==0 || ledpos==10) direction = -direction;
	}
}
